@@ -27,6 +27,7 @@ import (
 	"github.com/oklog/run"
 	"github.com/spf13/cobra"
 	"golang.org/x/oauth2"
+	"golang.org/x/oauth2/clientcredentials"
 
 	"github.com/openshift/telemeter/pkg/authorize"
 	"github.com/openshift/telemeter/pkg/authorize/jwt"
@@ -36,8 +37,9 @@ import (
 	telemeter_http "github.com/openshift/telemeter/pkg/http"
 	httpserver "github.com/openshift/telemeter/pkg/http/server"
 	"github.com/openshift/telemeter/pkg/metricfamily"
-	telemeter_oauth2 "github.com/openshift/telemeter/pkg/oauth2"
+	"github.com/openshift/telemeter/pkg/receive"
 	"github.com/openshift/telemeter/pkg/store"
+	"github.com/openshift/telemeter/pkg/store/forward"
 	"github.com/openshift/telemeter/pkg/store/memstore"
 	"github.com/openshift/telemeter/pkg/store/ratelimited"
 	"github.com/openshift/telemeter/pkg/validate"
@@ -109,15 +111,16 @@ func main() {
 	cmd.Flags().StringVar(&opt.SharedKey, "shared-key", opt.SharedKey, "The path to a private key file that will be used to sign authentication requests and secure the cluster protocol.")
 	cmd.Flags().Int64Var(&opt.TokenExpireSeconds, "token-expire-seconds", opt.TokenExpireSeconds, "The expiration of auth tokens in seconds.")
 
-	cmd.Flags().StringVar(&opt.AuthorizeEndpoint, "authorize", opt.AuthorizeEndpoint, "A endpoint URL to authorize against when a client requests a token.")
+	cmd.Flags().StringVar(&opt.AuthorizeEndpoint, "authorize", opt.AuthorizeEndpoint, "A URL against which to authorize client requests.")
 
-	cmd.Flags().StringVar(&opt.AuthorizeIssuerURL, "authorize-issuer-url", opt.AuthorizeIssuerURL, "The authorize OIDC issuer URL, see https://openid.net/specs/openid-connect-discovery-1_0.html#IssuerDiscovery.")
-	cmd.Flags().StringVar(&opt.AuthorizeUsername, "authorize-username", opt.AuthorizeUsername, "The authorize OIDC username, see rfc6749#section-4.3.")
-	cmd.Flags().StringVar(&opt.AuthorizePassword, "authorize-password", opt.AuthorizePassword, "The authorize OIDC password, see rfc6749#section-4.3.")
-	cmd.Flags().StringVar(&opt.AuthorizeClientID, "authorize-client-id", opt.AuthorizeClientID, "The authorize OIDC client ID, see rfc6749#section-4.3.")
+	cmd.Flags().StringVar(&opt.OIDCIssuer, "oidc-issuer", opt.OIDCIssuer, "The OIDC issuer URL, see https://openid.net/specs/openid-connect-discovery-1_0.html#IssuerDiscovery.")
+	cmd.Flags().StringVar(&opt.ClientSecret, "client-secret", opt.ClientSecret, "The OIDC client secret, see https://tools.ietf.org/html/rfc6749#section-2.3.")
+	cmd.Flags().StringVar(&opt.ClientID, "client-id", opt.ClientID, "The OIDC client ID, see https://tools.ietf.org/html/rfc6749#section-2.3.")
+	cmd.Flags().StringVar(&opt.TenantKey, "tenant-key", opt.TenantKey, "The JSON key in the bearer token whose value to use as the tenant ID.")
 
 	cmd.Flags().DurationVar(&opt.Ratelimit, "ratelimit", opt.Ratelimit, "The rate limit of metric uploads per cluster ID. Uploads happening more often than this limit will be rejected.")
 	cmd.Flags().DurationVar(&opt.TTL, "ttl", opt.TTL, "The TTL for metrics to be held in memory.")
+	cmd.Flags().StringVar(&opt.ForwardURL, "forward-url", opt.ForwardURL, "All written metrics will be written to this URL additionally")
 
 	cmd.Flags().BoolVarP(&opt.Verbose, "verbose", "v", opt.Verbose, "Show verbose output.")
 
@@ -150,10 +153,10 @@ type Options struct {
 
 	AuthorizeEndpoint string
 
-	AuthorizeIssuerURL string
-	AuthorizeClientID  string
-	AuthorizeUsername  string
-	AuthorizePassword  string
+	OIDCIssuer   string
+	ClientID     string
+	ClientSecret string
+	TenantKey    string
 
 	PartitionKey      string
 	LabelFlag         []string
@@ -165,8 +168,9 @@ type Options struct {
 	ElideLabels       []string
 	WhitelistFile     string
 
-	TTL       time.Duration
-	Ratelimit time.Duration
+	TTL        time.Duration
+	Ratelimit  time.Duration
+	ForwardURL string
 
 	Verbose bool
 }
@@ -232,8 +236,8 @@ func (o *Options) Run() error {
 			Transport: telemeter_http.NewInstrumentedRoundTripper("authorize", transport),
 		}
 
-		if o.AuthorizeIssuerURL != "" {
-			provider, err := oidc.NewProvider(ctx, o.AuthorizeIssuerURL)
+		if o.OIDCIssuer != "" {
+			provider, err := oidc.NewProvider(ctx, o.OIDCIssuer)
 			if err != nil {
 				return fmt.Errorf("OIDC provider initialization failed: %v", err)
 			}
@@ -245,22 +249,15 @@ func (o *Options) Run() error {
 				},
 			)
 
-			cfg := oauth2.Config{
-				ClientID: o.AuthorizeClientID,
-				Endpoint: provider.Endpoint(),
+			cfg := clientcredentials.Config{
+				ClientID:     o.ClientID,
+				ClientSecret: o.ClientSecret,
+				TokenURL:     provider.Endpoint().TokenURL,
 			}
 
-			src := telemeter_oauth2.NewPasswordCredentialsTokenSource(
-				ctx, &cfg,
-				o.AuthorizeUsername, o.AuthorizePassword,
-			)
-
-			// both the underlying upstream authorize transport
-			// and the oauth transport are already instrumented,
-			// hence this doesn't need another instrumentation.
 			authorizeClient.Transport = &oauth2.Transport{
 				Base:   authorizeClient.Transport,
-				Source: src,
+				Source: cfg.TokenSource(ctx),
 			}
 		}
 	}
@@ -361,9 +358,7 @@ func (o *Options) Run() error {
 	secret := h.Sum(nil)[:32]
 
 	external := http.NewServeMux()
-	externalProtected := http.NewServeMux()
 	internal := http.NewServeMux()
-	internalProtected := http.NewServeMux()
 
 	internalPaths := []string{"/", "/federate", "/metrics", "/debug/pprof", "/healthz", "/healthz/ready"}
 
@@ -374,13 +369,25 @@ func (o *Options) Run() error {
 	}
 
 	auth := jwt.NewAuthorizeClusterHandler(o.PartitionKey, o.TokenExpireSeconds, signer, o.RequiredLabels, clusterAuth)
-	validator := validate.New(o.PartitionKey, o.LimitBytes, 24*time.Hour)
+	validator := validate.New(o.PartitionKey, o.LimitBytes, 24*time.Hour, time.Now)
+
+	var store store.Store
 
 	ms := memstore.New(o.TTL)
 	ms.StartCleaner(ctx, time.Minute)
+	store = ms
+
+	// If specified all written metrics will be written to the remote forward URL
+	if o.ForwardURL != "" {
+		u, err := url.Parse(o.ForwardURL)
+		if err != nil {
+			return fmt.Errorf("--forward-url must be a valid URL: %v", err)
+		}
+		store = forward.New(u, store)
+	}
 
 	// Create a rate-limited store with a memory-store as its backend.
-	var store store.Store = ratelimited.New(o.Ratelimit, ms)
+	store = ratelimited.New(o.Ratelimit, store)
 
 	if len(o.ListenCluster) > 0 {
 		c := cluster.NewDynamic(o.Name, store)
@@ -404,7 +411,7 @@ func (o *Options) Run() error {
 			}()
 		}
 		internalPaths = append(internalPaths, "/debug/cluster")
-		internalProtected.Handle("/debug/cluster", c)
+		internal.Handle("/debug/cluster", c)
 		store = c
 		// Wrap the cluster store within a rate-limited store.
 		// This guarantees an upper-bound on the total inter-node requests that
@@ -426,13 +433,13 @@ func (o *Options) Run() error {
 	transforms.With(metricfamily.NewElide(o.ElideLabels...))
 
 	server := httpserver.New(store, validator, transforms, o.TTL)
+	receiver := receive.NewHandler(o.ForwardURL)
 
 	internalPathJSON, _ := json.MarshalIndent(Paths{Paths: internalPaths}, "", "  ")
-	externalPathJSON, _ := json.MarshalIndent(Paths{Paths: []string{"/", "/authorize", "/upload", "/healthz", "/healthz/ready"}}, "", "  ")
+	externalPathJSON, _ := json.MarshalIndent(Paths{Paths: []string{"/", "/authorize", "/upload", "/healthz", "/healthz/ready", "/metrics/v1/receive"}}, "", "  ")
 
 	// TODO: add internal authorization
-	telemeter_http.DebugRoutes(internalProtected)
-	internalProtected.Handle("/federate", http.HandlerFunc(server.Get))
+	telemeter_http.DebugRoutes(internal)
 
 	internal.Handle("/", http.HandlerFunc(func(w http.ResponseWriter, req *http.Request) {
 		if req.URL.Path == "/" && req.Method == "GET" {
@@ -442,14 +449,12 @@ func (o *Options) Run() error {
 			}
 			return
 		}
-		internalProtected.ServeHTTP(w, req)
+		w.WriteHeader(http.StatusNotFound)
 	}))
+	internal.Handle("/federate", http.HandlerFunc(server.Get))
 	telemeter_http.MetricRoutes(internal)
 	telemeter_http.HealthRoutes(internal)
 
-	externalProtected.Handle("/upload", telemeter_http.NewInstrumentedHandler("upload", http.HandlerFunc(server.Post)))
-	externalProtectedHandler := authorize.NewAuthorizeClientHandler(jwtAuthorizer, externalProtected)
-
 	external.Handle("/", http.HandlerFunc(func(w http.ResponseWriter, req *http.Request) {
 		if req.URL.Path == "/" && req.Method == "GET" {
 			w.Header().Add("Content-Type", "application/json")
@@ -458,10 +463,28 @@ func (o *Options) Run() error {
 			}
 			return
 		}
-		externalProtectedHandler.ServeHTTP(w, req)
+		w.WriteHeader(http.StatusNotFound)
 	}))
 	telemeter_http.HealthRoutes(external)
+
+	// v1 routes
 	external.Handle("/authorize", telemeter_http.NewInstrumentedHandler("authorize", auth))
+	external.Handle("/upload",
+		authorize.NewAuthorizeClientHandler(jwtAuthorizer,
+			telemeter_http.NewInstrumentedHandler("upload",
+				http.HandlerFunc(server.Post),
+			),
+		),
+	)
+
+	// v1 routes
+	external.Handle("/metris/v1/receive",
+		telemeter_http.NewInstrumentedHandler("receive",
+			authorize.NewHandler(authorizeClient, authorizeURL, o.TenantKey,
+				http.HandlerFunc(receiver.Receive),
+			),
+		),
+	)
 
 	log.Printf("Starting telemeter-server %s on %s (internal=%s, cluster=%s)", o.Name, o.Listen, o.ListenInternal, o.ListenCluster)
 
